"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.tryQueryParser = exports.tryHtmlParser = exports.queryParser = exports.htmlParser = exports.checkIfNothingRemains = exports.createQueueFromItems = exports.parsedQueryLexer = exports.parsedHtmlLexer = exports.parseLexer = void 0;
const html_1 = require("./html");
const lexers_1 = require("./lexers");
const selector_1 = require("./selector");
const utils_1 = require("./utils");
const parseLexer = (input, lexer, compress, initialMode = []) => {
    const acc = [];
    const lexers = Object.values(lexer);
    const lexerKeys = Object.keys(lexer);
    let parsedIndex = 0;
    const mode = initialMode;
    do {
        let matchedValue = "";
        const matchedLexer = lexers.findIndex((lexer) => {
            const matchesMode = (!lexer.mode && mode.length === 0) ||
                lexer.mode === mode[mode.length - 1];
            if (!matchesMode) {
                return false;
            }
            if (typeof lexer.value === "string") {
                const groupMatch = input.slice(parsedIndex, parsedIndex + lexer.value.length);
                if (groupMatch === lexer.value) {
                    parsedIndex += groupMatch.length;
                    matchedValue = groupMatch;
                }
                else {
                    return false;
                }
            }
            else if (typeof lexer.value === "function") {
                const executed = lexer.value(input, parsedIndex);
                if (typeof executed === "number") {
                    const groupMatch = input.slice(parsedIndex, executed);
                    parsedIndex += groupMatch.length;
                    matchedValue = groupMatch;
                }
                else {
                    return false;
                }
            }
            else {
                const matchesRegex = new RegExp(`^(${lexer.value.source})`, "gu").exec(input.slice(parsedIndex));
                if (!matchesRegex?.[0]) {
                    return false;
                }
                const groupMatch = matchesRegex[0];
                parsedIndex += groupMatch.length;
                matchedValue = groupMatch;
            }
            if (lexer.popMode) {
                mode.pop();
            }
            if (lexer.pushMode) {
                mode.push(...lexer.pushMode);
            }
            return true;
        });
        if (matchedLexer === -1) {
            throw `${input
                .slice(parsedIndex)
                .slice(0, 25)}... does not match any known lexer items, matched previously: ${acc
                .map(({ type, value }) => `${type}:"${value}"`)
                .slice(-10)
                .join(",")}`;
        }
        if (!compress || lexerKeys[matchedLexer] !== "SEA_WS") {
            acc.push({
                type: lexerKeys[matchedLexer],
                value: matchedValue,
            });
        }
    } while (parsedIndex < input.length);
    acc.push({ type: "EOF", value: "" });
    return {
        queue: acc,
        mode: mode,
    };
};
exports.parseLexer = parseLexer;
exports.parsedHtmlLexer = Object.entries(lexers_1.htmlLexerAtoms).reduce((lexer, [lexerKey, lexerValue]) => {
    const typedKey = lexerKey;
    lexer[typedKey] =
        lexerValue instanceof RegExp ||
            typeof lexerValue === "function" ||
            typeof lexerValue === "string"
            ? { value: lexerValue }
            : lexerValue;
    return lexer;
}, {});
exports.parsedQueryLexer = Object.entries(lexers_1.cssLexerAtoms).reduce((lexer, [lexerKey, lexerValue]) => {
    const typedKey = lexerKey;
    lexer[typedKey] =
        lexerValue instanceof RegExp || typeof lexerValue === "string"
            ? { value: lexerValue }
            : lexerValue;
    return lexer;
}, {});
const parseHtmlLexer = (input, compress) => {
    return (0, exports.parseLexer)(input, exports.parsedHtmlLexer, compress).queue;
};
const parseQueryLexer = (input) => {
    return (0, exports.parseLexer)(input, exports.parsedQueryLexer, false).queue;
};
const createQueueFromItems = (lexerAtoms) => {
    const createQueue = (at) => {
        return {
            items: lexerAtoms,
            at: at,
            next: () => createQueue(at + 1),
        };
    };
    return createQueue(0);
};
exports.createQueueFromItems = createQueueFromItems;
const checkIfNothingRemains = (lexerAtoms, endQueue, parent) => {
    const remainingLexerAtoms = lexerAtoms
        .slice(endQueue.at)
        .filter((atom) => atom.type !== "EOF");
    if (remainingLexerAtoms.length > 0) {
        throw `Error in parsing (finished: ${endQueue.at}/${lexerAtoms.length}) encountered at: ${remainingLexerAtoms
            .slice(0, 5)
            .map(({ type, value }) => `${type}: "${value}"`)
            .join(", ")}, remainder near error: ${(0, utils_1.parserItemToString)(parent).slice(-50)}`;
    }
};
exports.checkIfNothingRemains = checkIfNothingRemains;
const htmlParser = (input, compress) => {
    const lexerAtoms = parseHtmlLexer(input, compress);
    const document = new html_1.HtmlDocument();
    const endQueue = document.process((0, exports.createQueueFromItems)(lexerAtoms));
    (0, exports.checkIfNothingRemains)(lexerAtoms, endQueue, document);
    for (const d of document.descendants()) {
        d.getClassNames();
    }
    return document;
};
exports.htmlParser = htmlParser;
let queryCache = {};
let queryCount = 0;
const queryLimit = 50;
const queryParser = (input) => {
    if (queryCache[input]) {
        return queryCache[input];
    }
    const lexerAtoms = parseQueryLexer(input);
    const selectorGroup = new selector_1.SelectorGroup();
    const endQueue = selectorGroup.process((0, exports.createQueueFromItems)(lexerAtoms));
    (0, exports.checkIfNothingRemains)(lexerAtoms, endQueue, selectorGroup);
    if (queryCount > queryLimit) {
        queryCount = 0;
        queryCache = {};
    }
    queryCache[input] = selectorGroup;
    queryCount++;
    return selectorGroup;
};
exports.queryParser = queryParser;
const tryHtmlParser = (input, compress) => {
    try {
        return (0, exports.htmlParser)(input, compress);
    }
    catch {
        return undefined;
    }
};
exports.tryHtmlParser = tryHtmlParser;
const tryQueryParser = (input) => {
    try {
        return (0, exports.queryParser)(input);
    }
    catch {
        return undefined;
    }
};
exports.tryQueryParser = tryQueryParser;
